@@ -1,7 +1,7 @@
 /*
- * Copyright  2001 Keith Packard
+ * Copyright ® 2001 Keith Packard
  *
- * Partly based on code that is Copyright  The XFree86 Project Inc.
+ * Partly based on code that is Copyright ® The XFree86 Project Inc.
  *
  * Permission to use, copy, modify, distribute, and sell this software and its
  * documentation for any purpose is hereby granted without fee, provided that
@@ -23,7 +23,7 @@
  *
  * Authors:
  *    Eric Anholt <eric@anholt.net>
- *    Michel Dnzer <michel@tungstengraphics.com>
+ *    Michel Dänzer <michel@tungstengraphics.com>
  *
  */
 
@@ -188,7 +188,8 @@ uxa_do_put_image (DrawablePtr pDrawable, GCPtr pGC, int depth, int x, int y,
 	    int	dstXoff, dstYoff;
 
 	    if (!access_prepared) {
-		uxa_prepare_access(pDrawable, UXA_ACCESS_RW);
+		if (!uxa_prepare_access(pDrawable, UXA_ACCESS_RW))
+		    return FALSE;
 		access_prepared = TRUE;
 	    }
 
@@ -215,6 +216,8 @@ uxa_do_put_image (DrawablePtr pDrawable, GCPtr pGC, int depth, int x, int y,
 
 #ifdef MITSHM
 
+#include "xorgVersion.h"
+
 static Bool
 uxa_do_shm_put_image(DrawablePtr pDrawable, GCPtr pGC, int depth,
 		 unsigned int format, int w, int h, int sx, int sy, int sw,
@@ -237,7 +240,8 @@ uxa_do_shm_put_image(DrawablePtr pDrawable, GCPtr pGC, int depth,
 	if (!pPixmap)
 	    return FALSE;
 
-        uxa_prepare_access (pDrawable, UXA_ACCESS_RW);
+        if (!uxa_prepare_access (pDrawable, UXA_ACCESS_RW))
+	    return FALSE;
 	fbCopyArea((DrawablePtr)pPixmap, pDrawable, pGC, sx, sy, sw, sh, dx, dy);
 	uxa_finish_access(pDrawable);
 
@@ -249,6 +253,8 @@ uxa_do_shm_put_image(DrawablePtr pDrawable, GCPtr pGC, int depth,
     return FALSE;
 }
 
+#if XORG_VERSION_CURRENT < XORG_VERSION_NUMERIC(1,5,99,0,0)
+
 /* The actual ShmPutImage isn't wrapped by the damage layer, so we need to
  * inform any interested parties of the damage incurred to the drawable.
  *
@@ -262,12 +268,16 @@ uxa_shm_put_image(DrawablePtr pDrawable, GCPtr pGC, int depth, unsigned int form
 {
     if (!uxa_do_shm_put_image(pDrawable, pGC, depth, format, w, h, sx, sy, sw, sh,
 			      dx, dy, data)) {
-	uxa_prepare_access (pDrawable, UXA_ACCESS_RW);
+	if (!uxa_prepare_access (pDrawable, UXA_ACCESS_RW))
+	    return;
 	fbShmPutImage(pDrawable, pGC, depth, format, w, h, sx, sy, sw, sh, dx, dy,
 		      data);
 	uxa_finish_access(pDrawable);
     }
 }
+#else
+#define uxa_shm_put_image NULL
+#endif
 
 ShmFuncs uxa_shm_funcs = { NULL, uxa_shm_put_image };
 
@@ -421,7 +431,8 @@ uxa_copy_n_to_n (DrawablePtr    pSrcDrawable,
 		 Pixel		bitplane,
 		 void		*closure)
 {
-    uxa_screen_t    *uxa_screen = uxa_get_screen(pDstDrawable->pScreen);
+    ScreenPtr       screen = pDstDrawable->pScreen;
+    uxa_screen_t    *uxa_screen = uxa_get_screen(screen);
     int		    src_off_x, src_off_y;
     int		    dst_off_x, dst_off_y;
     PixmapPtr	    pSrcPixmap, pDstPixmap;
@@ -468,12 +479,14 @@ fallback:
     UXA_FALLBACK(("from %p to %p (%c,%c)\n", pSrcDrawable, pDstDrawable,
 		  uxa_drawable_location(pSrcDrawable),
 		  uxa_drawable_location(pDstDrawable)));
-    uxa_prepare_access (pDstDrawable, UXA_ACCESS_RW);
-    uxa_prepare_access (pSrcDrawable, UXA_ACCESS_RO);
-    fbCopyNtoN (pSrcDrawable, pDstDrawable, pGC, pbox, nbox, dx, dy, reverse,
-		upsidedown, bitplane, closure);
-    uxa_finish_access (pSrcDrawable);
-    uxa_finish_access (pDstDrawable);
+    if (uxa_prepare_access (pDstDrawable, UXA_ACCESS_RW)) {
+	if (uxa_prepare_access (pSrcDrawable, UXA_ACCESS_RO)) {
+	    fbCopyNtoN (pSrcDrawable, pDstDrawable, pGC, pbox, nbox, dx, dy,
+			reverse, upsidedown, bitplane, closure);
+	    uxa_finish_access (pSrcDrawable);
+	}
+	uxa_finish_access (pDstDrawable);
+    }
 }
 
 RegionPtr
@@ -487,9 +500,9 @@ uxa_copy_area(DrawablePtr pSrcDrawable, DrawablePtr pDstDrawable, GCPtr pGC,
                                  srcx, srcy, width, height, dstx, dsty);
     }
 
-    return  fbDoCopy (pSrcDrawable, pDstDrawable, pGC,
-                      srcx, srcy, width, height,
-                      dstx, dsty, uxa_copy_n_to_n, 0, NULL);
+    return miDoCopy (pSrcDrawable, pDstDrawable, pGC,
+		     srcx, srcy, width, height,
+		     dstx, dsty, uxa_copy_n_to_n, 0, NULL);
 }
 
 static void
@@ -835,7 +848,7 @@ uxa_copy_window(WindowPtr pWin, DDXPointRec ptOldOrg, RegionPtr prgnSrc)
 			  -pPixmap->screen_x, -pPixmap->screen_y);
 #endif
 
-    fbCopyRegion (&pPixmap->drawable, &pPixmap->drawable,
+    miCopyRegion (&pPixmap->drawable, &pPixmap->drawable,
 		  NULL,
 		  &rgnDst, dx, dy, uxa_copy_n_to_n, 0, NULL);
 
@@ -971,7 +984,6 @@ out:
     return ret;
 }
 
-
 /**
  * Accelerates GetImage for solid ZPixmap downloads from framebuffer memory.
  *
@@ -983,7 +995,8 @@ void
 uxa_get_image (DrawablePtr pDrawable, int x, int y, int w, int h,
 	       unsigned int format, unsigned long planeMask, char *d)
 {
-    uxa_screen_t    *uxa_screen = uxa_get_screen(pDrawable->pScreen);
+    ScreenPtr       screen = pDrawable->pScreen;
+    uxa_screen_t    *uxa_screen = uxa_get_screen(screen);
     BoxRec	    Box;
     PixmapPtr	    pPix = uxa_get_drawable_pixmap (pDrawable);
     int		    xoff, yoff;
@@ -1024,9 +1037,10 @@ fallback:
     UXA_FALLBACK(("from %p (%c)\n", pDrawable,
 		  uxa_drawable_location(pDrawable)));
 
-    uxa_prepare_access (pDrawable, UXA_ACCESS_RO);
-    fbGetImage (pDrawable, x, y, w, h, format, planeMask, d);
-    uxa_finish_access (pDrawable);
+    if (uxa_prepare_access (pDrawable, UXA_ACCESS_RO)) {
+	fbGetImage (pDrawable, x, y, w, h, format, planeMask, d);
+	uxa_finish_access (pDrawable);
+    }
 
    return;
 }
